@@ -21,6 +21,7 @@
 #
 
 require 'rexml/parsers/baseparser'
+require 'kramdown/parser/html'
 
 module Kramdown
 
@@ -50,7 +51,7 @@ module Kramdown
       end
 
       include ::Kramdown::Utils::Html
-
+      include ::Kramdown::Parser::Html::Constants
 
       # The amount of indentation used when nesting HTML tags.
       attr_accessor :indent
@@ -64,6 +65,7 @@ module Kramdown
         @toc_code = nil
         @indent = 2
         @stack = []
+        @coderay_enabled = @options[:enable_coderay] && HIGHLIGHTING_AVAILABLE
       end
 
       # The mapping of element type to conversion method.
@@ -108,12 +110,13 @@ module Kramdown
       end
 
       def convert_codeblock(el, indent)
-        if (el.attr['lang'] || @options[:coderay_default_lang]) && HIGHLIGHTING_AVAILABLE
-          attr = el.attr.dup
+        attr = el.attr.dup
+        lang = extract_code_language!(attr)
+        if @coderay_enabled && (lang || @options[:coderay_default_lang])
           opts = {:wrap => @options[:coderay_wrap], :line_numbers => @options[:coderay_line_numbers],
             :line_number_start => @options[:coderay_line_number_start], :tab_width => @options[:coderay_tab_width],
             :bold_every => @options[:coderay_bold_every], :css => @options[:coderay_css]}
-          lang = (attr.delete('lang') || @options[:coderay_default_lang]).to_sym
+          lang = (lang || @options[:coderay_default_lang]).to_sym
           result = CodeRay.scan(el.value, lang).html(opts).chomp << "\n"
           "#{' '*indent}<div#{html_attributes(attr)}>#{result}#{' '*indent}</div>\n"
         else
@@ -130,7 +133,9 @@ module Kramdown
               end.join('')
             end
           end
-          "#{' '*indent}<pre#{html_attributes(el.attr)}><code>#{result}\n</code></pre>\n"
+          code_attr = {}
+          code_attr['class'] = "language-#{lang}" if lang
+          "#{' '*indent}<pre#{html_attributes(attr)}><code#{html_attributes(code_attr)}>#{result}\n</code></pre>\n"
         end
       end
 
@@ -144,7 +149,8 @@ module Kramdown
           attr['id'] = generate_id(el.options[:raw_text])
         end
         @toc << [el.options[:level], attr['id'], el.children] if attr['id'] && in_toc?(el)
-        "#{' '*indent}<h#{el.options[:level]}#{html_attributes(attr)}>#{inner(el, indent)}</h#{el.options[:level]}>\n"
+        level = output_header_level(el.options[:level])
+        "#{' '*indent}<h#{level}#{html_attributes(attr)}>#{inner(el, indent)}</h#{level}>\n"
       end
 
       def convert_hr(el, indent)
@@ -178,25 +184,24 @@ module Kramdown
         "#{' '*indent}<dt#{html_attributes(el.attr)}>#{inner(el, indent)}</dt>\n"
       end
 
-      # A list of all HTML tags that need to have a body (even if the body is empty).
-      HTML_TAGS_WITH_BODY=['div', 'span', 'script', 'iframe', 'textarea', 'a'] # :nodoc:
-
       def convert_html_element(el, indent)
         res = inner(el, indent)
         if el.options[:category] == :span
-          "<#{el.value}#{html_attributes(el.attr)}" << (!res.empty? || HTML_TAGS_WITH_BODY.include?(el.value) ? ">#{res}</#{el.value}>" : " />")
+          "<#{el.value}#{html_attributes(el.attr)}" << (res.empty? && HTML_ELEMENTS_WITHOUT_BODY.include?(el.value) ? " />" : ">#{res}</#{el.value}>")
         else
           output = ''
           output << ' '*indent if @stack.last.type != :html_element || @stack.last.options[:content_model] != :raw
           output << "<#{el.value}#{html_attributes(el.attr)}"
-          if !res.empty? && el.options[:content_model] != :block
+          if el.options[:is_closed] && el.options[:content_model] == :raw
+            output << " />"
+          elsif !res.empty? && el.options[:content_model] != :block
             output << ">#{res}</#{el.value}>"
           elsif !res.empty?
             output << ">\n#{res.chomp}\n"  << ' '*indent << "</#{el.value}>"
-          elsif HTML_TAGS_WITH_BODY.include?(el.value)
-            output << "></#{el.value}>"
-          else
+          elsif HTML_ELEMENTS_WITHOUT_BODY.include?(el.value)
             output << " />"
+          else
+            output << "></#{el.value}>"
           end
           output << "\n" if @stack.last.type != :html_element || @stack.last.options[:content_model] != :raw
           output
@@ -253,8 +258,9 @@ module Kramdown
         res = inner(el, indent)
         attr = el.attr.dup
         if attr['href'] =~ /^mailto:/
-          attr['href'] = obfuscate('mailto') << ":" << obfuscate(attr['href'].sub(/^mailto:/, ''))
-          res = obfuscate(res)
+          mail_addr = attr['href'].sub(/^mailto:/, '')
+          attr['href'] = obfuscate('mailto') << ":" << obfuscate(mail_addr)
+          res = obfuscate(res) if res == mail_addr
         end
         "<a#{html_attributes(attr)}>#{res}</a>"
       end
@@ -264,10 +270,10 @@ module Kramdown
       end
 
       def convert_codespan(el, indent)
-        if el.attr['lang'] && HIGHLIGHTING_AVAILABLE
-          attr = el.attr.dup
-          result = CodeRay.scan(el.value, attr.delete('lang').to_sym).html(:wrap => :span, :css => @options[:coderay_css]).chomp
-          "<code#{html_attributes(attr)}>#{result}</code>"
+        lang = extract_code_language(el.attr)
+        if @coderay_enabled && lang
+          result = CodeRay.scan(el.value, lang.to_sym).html(:wrap => :span, :css => @options[:coderay_css]).chomp
+          "<code#{html_attributes(el.attr)}>#{result}</code>"
         else
           "<code#{html_attributes(el.attr)}>#{escape_html(el.value)}</code>"
         end
@@ -322,7 +328,7 @@ module Kramdown
 
       def convert_abbreviation(el, indent)
         title = @root.options[:abbrev_defs][el.value]
-        "<abbr#{!title.empty? ? " title=\"#{title}\"" : ''}>#{el.value}</abbr>"
+        "<abbr#{!title.empty? ? html_attributes(:title => title) : ''}>#{el.value}</abbr>"
       end
 
       def convert_root(el, indent)
@@ -390,7 +396,7 @@ module Kramdown
         text.each_byte do |b|
           result << (b > 128 ? b.chr : "&#%03d;" % b)
         end
-        result.force_encoding(text.encoding) if RUBY_VERSION >= '1.9'
+        result.force_encoding(text.encoding) if result.respond_to?(:force_encoding)
         result
       end
 
